load "Recognition.m";

function SubmatrixOfRank(P, TargetRank, ColumnsOrRows);
// Input:   A period matrix P of dimension r x c along with a tolerance epsinv
//          for testing triviality of subdeterminants.
//          The number of rows/columns to extract, whether to extract rows or
//		 columns.
// Output:  A TargetRank x #columns submatrix M and the corresponding subset 
//          of {1..#rows/#columns}. The matrix M should have rank TargetRank
//          We do not attempt to find the best one.


Flipped := false;
if ColumnsOrRows eq "Columns" then
    Flipped := true;
    P := Transpose(P);
end if;

r := #Rows(P);
c := #Rows(Transpose(P));


RP := Rows(P);
for s in Subsets({1..r}, TargetRank) do
    s0 := s;
    P0 := Matrix([RP[i] : i in s]);
    if NumericalRank(P0) eq TargetRank then 
        if not Flipped then
            return P0, s0;
        else
            return Transpose(P0), s0;
        end if;
    end if;
end for;

error Error("Failed to find submatrix with the desired rank");

end function;

function MatrixRatio(M1,M2);
    for i in [1..4] do
        for j in [1..4] do
            if M2[i][j] ne 0 then
                return M1[i][j]/M2[i][j];
            end if;
        end for;
    end for;
end function;



function BuildAMatrices(B, P);
	A := [];
	C := Parent(P[1][1]);
	for b in B do
		RHS := P*ChangeRing(b,C);
		FullRankSubm, cols := SubmatrixOfRank(P, 2, "Columns");
		RHS := Matrix(C, 2,2, [ RHS[i][j] : j in cols, i in [1..2] ]);
		a := RHS * FullRankSubm^(-1);
		A := Append(A,a);		
	end for;
	return A;
end function;


function IsProjection(M);
	if M eq ZeroMatrix(Integers(),4,4) then
		return false;
	end if;
	ratio:=MatrixRatio(M^2,M);
	return M^2 eq ratio*M;
end function;



function IsRationalProjection(M,A);
	if IsProjection(M) then
		if RecognizeNumberField(A) eq Rationals() then
			return true, Abs(MatrixRatio(M^2,M));
		end if;
	end if;
	return false, _;
end function;


function FindProjector2(endo, alpha, P);
	b := Basis(endo);
	n := #b;
	t := CartesianPower([-1..1], n);
	C := Parent(P[1][1]);
	
	A := BuildAMatrices(b, P);

	MinDegree := 10^3;

	for c in t do
		m := &+ [ b[i] * c[i] : i in [1..n] ] ;
		a := &+ [ A[i] * c[i] : i in [1..n] ] ;

		ev:=Eigenvalues(m);
		if not IsEmpty(ev) then
			eigVal:={@ z : z in ev@};
			TargetEigenvalue:=eigVal[1][1];
	
			MMatrix := m - TargetEigenvalue;
			AMatrix := a - TargetEigenvalue;

			proj, deg := IsRationalProjection(MMatrix,AMatrix);
			if proj then
				if deg lt MinDegree then
					MinDegree := deg;
					ReturnA := AMatrix;
					ReturnM := MMatrix;
				end if;
			end if;
		end if;
	end for;
	if MinDegree eq 10^3 then
		return false, _, _, _;
	else
		return true, ReturnM, ReturnA, Integers()!MinDegree;
	end if;
end function;




function ExtractLatticeGenerators(M);
// Input: a 1 x n matrix with entries in C whose "columns" should span a rank-2 lattice
// Output: a 1 x 2 matrix that generates the same lattice

g1:=0;
g2:=0;

C := BaseRing(M); // recover the complex field containing the coefficients of M
eps := 10^(-4); // numbers smaller than eps in absolute values are rounded to 0 to decide linear dependence
n := #Rows(Transpose(M));

// Loop through the coefficients of M to find the first nonzero entry
for i in [1..n] do
    if Abs(M[1][i]) ge eps then
        g1 := M[1][i];
        break i;
    end if;
end for;

if g1 eq 0 then
	return Matrix(C,1,2,[0,0]), false;
end if;

// Loop through the coefficients of M to find an entry that is not RR-linearly dependent from g1
for i in [1..n] do
    if Abs(M[1][i]) ge eps and Abs(Im(M[1][i]/g1)) ge eps then
        g2 := M[1][i];
        break i;
    end if;
end for;

// the rational span of g1, g2 should be the same as the rational span of all the entries of M.
// Now loop through the entries of M once again and express them as rational linear combinations of g1, g2

CoordinateVectors:=Matrix(Rationals(),2,n,[]);
for i in [1..n] do
        FindLinearRelation := [g1,g2,M[1][i]];
	CoefficientsLinearRelation:=LinearRelation(FindLinearRelation);
    if CoefficientsLinearRelation[3] ne 0 then
CoordinateVectors[1][i]:=-CoefficientsLinearRelation[1]/CoefficientsLinearRelation[3];
CoordinateVectors[2][i]:=-CoefficientsLinearRelation[2]/CoefficientsLinearRelation[3];
    end if;
end for;

// Let's recover the lattice itself
L:=Lattice(Transpose(CoordinateVectors));
// Find a basis of the lattice
B:=Basis(L);

// h1, h2 should now be complex numbers that generate the same lattice as all the entries of M
h1:=B[1][1]*g1+B[1][2]*g2;
h2:=B[2][1]*g1+B[2][2]*g2;

// return h1, h2 in matrix form
return Matrix(C,1,2,[h1,h2]), true;
end function;
